@@ -35,7 +35,9 @@
 #  include <sys/un.h>
 #  include <netinet/in.h>
 #  include <arpa/inet.h>
+#  include <resolv.h>
 #  include <netdb.h>
+#  include <sys/stat.h>
 #endif /* G_OS_WIN32 */
 #include <unistd.h>
 #include <stdio.h>
@@ -111,6 +113,7 @@ struct _SockSource {
 static guint io_timeout = 60;
 
 static GList *sock_connect_data_list = NULL;
+static GList *sock_list = NULL;
 
 static gboolean sock_prepare		(GSource	*source,
 					 gint		*timeout);
@@ -126,6 +129,8 @@ GSourceFuncs sock_watch_funcs = {
 	NULL
 };
 
+static SockInfo *sock_find_from_fd	(gint	fd);
+
 static gint sock_connect_with_timeout	(gint			 sock,
 					 const struct sockaddr	*serv_addr,
 					 gint			 addrlen,
@@ -345,9 +350,40 @@ gint fd_accept(gint sock)
 	return accept(sock, (struct sockaddr *)&caddr, &caddr_len);
 }
 
+
+static SockInfo *sock_find_from_fd(gint fd)
+{
+	GList *cur;
+
+	for (cur = sock_list; cur != NULL; cur = cur->next) {
+		if (((SockInfo *)cur->data)->sock == fd)
+			return (SockInfo *)cur->data;
+	}
+
+	return NULL;
+}
+
 static gint set_nonblocking_mode(gint fd, gboolean nonblock)
 {
-#ifdef G_OS_UNIX
+#ifdef G_OS_WIN32
+	gulong val = nonblock ? 1 : 0;
+	SockInfo *sock;
+
+	if (!nonblock)
+		WSAEventSelect(fd, NULL, 0);
+	if (ioctlsocket(fd, FIONBIO, &val) == SOCKET_ERROR) {
+		g_warning("set_nonblocking_mode(): ioctlsocket() failed: %ld\n",
+			  WSAGetLastError());
+		return -1;
+	}
+
+	sock = sock_find_from_fd(fd);
+	if (sock)
+		sock->nonblock = nonblock;
+	debug_print("set nonblocking mode to %d\n", nonblock);
+
+	return 0;
+#else
 	gint flags;
 
 	flags = fcntl(fd, F_GETFL, 0);
@@ -362,21 +398,33 @@ static gint set_nonblocking_mode(gint fd, gboolean nonblock)
 		flags &= ~O_NONBLOCK;
 
 	return fcntl(fd, F_SETFL, flags);
-#else
-	return -1;
 #endif
 }
 
 gint sock_set_nonblocking_mode(SockInfo *sock, gboolean nonblock)
 {
+	gint ret;
+
 	g_return_val_if_fail(sock != NULL, -1);
 
-	return set_nonblocking_mode(sock->sock, nonblock);
+	ret = set_nonblocking_mode(sock->sock, nonblock);
+	if (ret == 0)
+		sock->nonblock = nonblock;
+
+	return ret;
 }
 
 static gboolean is_nonblocking_mode(gint fd)
 {
-#ifdef G_OS_UNIX
+#ifdef G_OS_WIN32
+	SockInfo *sock;
+
+	sock = sock_find_from_fd(fd);
+	if (sock)
+		return sock->nonblock;
+
+	return FALSE;
+#else
 	gint flags;
 
 	flags = fcntl(fd, F_GETFL, 0);
@@ -386,8 +434,6 @@ static gboolean is_nonblocking_mode(gint fd)
 	}
 
 	return ((flags & O_NONBLOCK) != 0);
-#else
-	return FALSE;
 #endif
 }
 
@@ -395,7 +441,11 @@ gboolean sock_is_nonblocking_mode(SockInfo *sock)
 {
 	g_return_val_if_fail(sock != NULL, FALSE);
 
+#ifdef G_OS_WIN32
+	return sock->nonblock;
+#else
 	return is_nonblocking_mode(sock->sock);
+#endif
 }
 
 gboolean sock_has_read_data(SockInfo *sock)
@@ -407,8 +457,11 @@ gboolean sock_has_read_data(SockInfo *sock)
 	if (sock->ssl)
 		return TRUE;
 #endif
-	if (ioctlsocket(sock->sock, FIONREAD, &val) < 0)
+	if (ioctlsocket(sock->sock, FIONREAD, &val) < 0) {
+		g_warning("sock_has_read_data(): ioctlsocket() failed: %ld\n",
+			  WSAGetLastError());
 		return TRUE;
+	}
 
 	if (val == 0)
 		return FALSE;
@@ -570,6 +623,21 @@ static gint sock_connect_with_timeout(gint sock,
 	return ret;
 }
 
+static void resolver_init(void)
+{
+#ifdef G_OS_UNIX
+	static time_t resolv_conf_mtime = 0;
+	struct stat s;
+
+	if (g_stat("/etc/resolv.conf", &s) == 0 &&
+	    s.st_mtime != resolv_conf_mtime) {
+		debug_print("Reloading /etc/resolv.conf\n");
+		resolv_conf_mtime = s.st_mtime;
+		res_init();
+	}
+#endif
+}
+
 struct hostent *my_gethostbyname(const gchar *hostname)
 {
 	struct hostent *hp;
@@ -633,6 +701,8 @@ static gint sock_connect_by_hostname(gint sock, const gchar *hostname,
 	struct hostent *hp;
 	struct sockaddr_in ad;
 
+	resolver_init();
+
 	memset(&ad, 0, sizeof(ad));
 	ad.sin_family = AF_INET;
 	ad.sin_port = htons(port);
@@ -678,6 +748,8 @@ static SockDesc sock_connect_by_getaddrinfo(const gchar *hostname, gushort port)
 	struct addrinfo hints, *res, *ai;
 	gchar port_str[6];
 
+	resolver_init();
+
 	memset(&hints, 0, sizeof(hints));
 	/* hints.ai_flags = AI_CANONNAME; */
 	hints.ai_family = AF_UNSPEC;
@@ -748,6 +820,9 @@ SockInfo *sock_connect(const gchar *hostname, gushort port)
 	sockinfo->hostname = g_strdup(hostname);
 	sockinfo->port = port;
 	sockinfo->state = CONN_ESTABLISHED;
+	sockinfo->nonblock = FALSE;
+
+	sock_list = g_list_prepend(sock_list, sockinfo);
 
 	g_usleep(100000);
 
@@ -805,6 +880,9 @@ static gboolean sock_connect_async_cb(GIOChannel *source,
 	sockinfo->hostname = g_strdup(conn_data->hostname);
 	sockinfo->port = conn_data->port;
 	sockinfo->state = CONN_ESTABLISHED;
+	sockinfo->nonblock = TRUE;
+
+	sock_list = g_list_prepend(sock_list, sockinfo);
 
 	conn_data->func(sockinfo, conn_data->data);
 
@@ -1037,6 +1115,8 @@ static SockLookupData *sock_get_address_info_async(const gchar *hostname,
 	gint pipe_fds[2];
 	pid_t pid;
 
+	resolver_init();
+
 	if (pipe(pipe_fds) < 0) {
 		perror("pipe");
 		func(NULL, data);
@@ -1177,6 +1257,21 @@ gint sock_printf(SockInfo *sock, const gchar *format, ...)
 	return sock_write_all(sock, buf, strlen(buf));
 }
 
+#ifdef G_OS_WIN32
+static void sock_set_errno_from_last_error(gint error)
+{
+	switch (error) {
+	case WSAEWOULDBLOCK:
+		errno = EAGAIN;
+		break;
+	default:
+		debug_print("last error = %d\n", error);
+		errno = 0;
+		break;
+	}
+}
+#endif
+
 gint sock_read(SockInfo *sock, gchar *buf, gint len)
 {
 	g_return_val_if_fail(sock != NULL, -1);
@@ -1190,12 +1285,12 @@ gint sock_read(SockInfo *sock, gchar *buf, gint len)
 
 gint fd_read(gint fd, gchar *buf, gint len)
 {
+#ifdef G_OS_WIN32
+	return fd_recv(fd, buf, len, 0);
+#else
 	if (fd_check_io(fd, G_IO_IN) < 0)
 		return -1;
 
-#ifdef G_OS_WIN32
-	return recv(fd, buf, len, 0);
-#else
 	return read(fd, buf, len);
 #endif
 }
@@ -1243,11 +1338,23 @@ gint sock_write(SockInfo *sock, const gchar *buf, gint len)
 
 gint fd_write(gint fd, const gchar *buf, gint len)
 {
+#ifdef G_OS_WIN32
+	gint ret;
+#endif
 	if (fd_check_io(fd, G_IO_OUT) < 0)
 		return -1;
 
 #ifdef G_OS_WIN32
-	return send(fd, buf, len, 0);
+	ret = send(fd, buf, len, 0);
+	if (ret == SOCKET_ERROR) {
+		gint err;
+		err = WSAGetLastError();
+		sock_set_errno_from_last_error(err);
+		if (err != WSAEWOULDBLOCK)
+			g_warning("fd_write() failed with %d (errno = %d)\n",
+				  err, errno);
+	}
+	return ret;
 #else
 	return write(fd, buf, len);
 #endif
@@ -1320,10 +1427,26 @@ gint ssl_write_all(SSL *ssl, const gchar *buf, gint len)
 
 gint fd_recv(gint fd, gchar *buf, gint len, gint flags)
 {
+#ifdef G_OS_WIN32
+	gint ret;
+#endif
 	if (fd_check_io(fd, G_IO_IN) < 0)
 		return -1;
 
+#ifdef G_OS_WIN32
+	ret = recv(fd, buf, len, flags);
+	if (ret == SOCKET_ERROR) {
+		gint err;
+		err = WSAGetLastError();
+		sock_set_errno_from_last_error(err);
+		if (err != WSAEWOULDBLOCK)
+			g_warning("fd_recv(): failed with %d (errno = %d)\n",
+				  err, errno);
+	}
+	return ret;
+#else
 	return recv(fd, buf, len, flags);
+#endif
 }
 
 gint fd_gets(gint fd, gchar *buf, gint len)
@@ -1500,6 +1623,8 @@ gint sock_peek(SockInfo *sock, gchar *buf, gint len)
 
 gint sock_close(SockInfo *sock)
 {
+	GList *cur;
+
 	if (!sock)
 		return 0;
 
@@ -1513,6 +1638,13 @@ gint sock_close(SockInfo *sock)
 		g_io_channel_unref(sock->sock_ch);
 	}
 
+	for (cur = sock_list; cur != NULL; cur = cur->next) {
+		if ((SockInfo *)cur->data == sock) {
+			sock_list = g_list_remove(sock_list, sock);
+			break;
+		}
+	}
+
 	g_free(sock->hostname);
 	g_free(sock);
 
